// Larbin
// Sebastien Ailleret
// 14-12-99 -> 11-09-01

#include <unistd.h>
#include <iostream.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "options.h"

#include "types.h"
#include "global.h"
#include "utils/text.h"
#include "utils/string.h"
#include "utils/Vector.h"
#include "fetch/site.h"
#include "fetch/file.h"
#include "fetch/fetchOpen.h"
#include "fetch/checker.h"

#include "utils/debug.h"

#define ANSWER 0
#define HEADERS 1
#define HEADERS30X 2
#define HTML 3
#define COMMENT 4
#define TAG 5
#define TAGCONTENT 6
#define SPECIFIC 7

#define NONE 0
#define AHREF 1
#define BASE 2
#define FRAME 3
//#define COMMENT 4    // ==> do not redefine 4

/***********************************
 * implementation of file
 ***********************************/

file::file (Connexion *conn) {
  buffer = conn->buffer;
  pos = 0;
  posParse = buffer;
}

file::~file () {
}

/** Is it a robots.txt */
bool file::isRobots () {
  return true;
}

/***********************************
 * implementation of robots
 ***********************************/

/** Constructor
 */
robots::robots (NamedSite *server, Connexion *conn) : file(conn) {
  newPars();
  this->server = server;
  answerCode = false;
}

/** Destructor
 */
robots::~robots () {
  delPars();
  // server is not deleted on purpose
  // it belongs to someone else
}

/** we get some more chars of this file
 */
int robots::endInput () {
  return 0;
}

/** input and parse headers
 */
int robots::inputHeaders (int size) {
  pos += size;
  if (!answerCode && pos > 12) {
    if (buffer[9] == '2') {
      answerCode = true;
    } else {
      errno = err40X;
      return 1;
    }
  }
  if (pos > maxRobotsSize) {
	// no more input, forget the end of this file
	errno = tooBig;
	return 1;
  } else {
	return 0;
  }
}

/** parse the robots.txt
 */
void robots::parse (bool isError) {
  if (answerCode && parseHeaders()) {
    siteRobots();
    buffer[pos] = 0;
	if (isError) {
	  // The file could be incomplete, delete last token
	  // We could have Disallow / instead of Disallow /blabla
	  for (uint i=pos-1; i>0 && !isspace(buffer[i]); i--) {
		buffer[i] = ' ';
	  }
	}
    uncomment();
    parseRobots();
  }
}

/** Uncomment this file
 */
void robots::uncomment () {
  bool comment = false;
  for (char *i=posParse; ; i++) {
    switch (*i) {
    case ':':
      *i = ' ';
      break;
    case '#':
      comment = 1;
      *i = ' ';
      break;
    case '\n':
      comment = 0;
      break;
    case 0:
      return;
    default:
      if (comment) {
        *i = ' ';
      }
      break;
    }
  }
}

/** test http headers
 * return true if OK, false otherwise
 */
bool robots::parseHeaders () {
  for(posParse = buffer+9; posParse[3] != 0; posParse++) {
    if ((posParse[0] == '\n' && 
         (posParse[1] == '\n'
          || posParse[2] == '\n'))
        || (posParse[0] == '\r' &&
            (posParse[1] == '\r'
             || posParse[2] == '\r'))) {
      return true;
    }
  }
  return false;
}

/** try to understand the file
 */
void robots::parseRobots () {
  robotsOK();
#ifndef NOSTATS
  bool goodfile = true;
#endif // NOSTATS
  assert (server->forbidden == NULL);
  server->forbidden = new Vector<char>;
  uint items = 0; // size of server->forbidden
  // state
  // 0 : not concerned
  // 1 : weakly concerned
  // 2 : strongly concerned
  int state = 1;
  char *tok = nextToken(&posParse);
  while (tok != NULL) {
	if (!strcasecmp(tok, "useragent") || !strcasecmp(tok, "user-agent")) {
	  if (state == 2) {
		// end of strong concern record => the end for us
		return;
	  } else {
		state = 0;
		// what is the new state ?
		tok = nextToken(&posParse);
		while (tok != NULL
			   && strcasecmp(tok, "useragent")
			   && strcasecmp(tok, "user-agent")
			   && strcasecmp(tok, "disallow")) {
          if (caseContain(tok, global::userAgent)) {
            state = 2;
          } else if (state == 0 && !strcmp(tok, "*")) {
            state = 1;
          }
		  tok = nextToken(&posParse);
		}
	  }
	  if (state) {
		// delete old forbidden : we've got a better record than older ones
		server->forbidden->recycle();
		items = 0;
	  } else {
        // forget this record
        while (tok != NULL
			   && strcasecmp(tok, "useragent")
			   && strcasecmp(tok, "user-agent")) {
          tok = nextToken(&posParse);
        }
      }
	} else if (!strcasecmp(tok, "disallow")) {
      tok = nextToken(&posParse);
      while (tok != NULL
             && strcasecmp(tok, "useragent")
             && strcasecmp(tok, "user-agent")
             && strcasecmp(tok, "disallow")) {
        // add nextToken to forbidden
        if (items++ < maxRobotsItem) {
          // allocate a correct token
          tok = makeGoodToken(tok);
          server->forbidden->addElement(tok);
        }
        tok = nextToken(&posParse);
      }
	} else {
#ifndef NOSTATS
	  if (goodfile) {
        robotsOKdec();
		goodfile = false;
	  }
#endif // NOSTATS
	  tok = nextToken(&posParse);
	}
  }
}

/** Given a token, allocate a good one corresponding
 * try to understand it as a path, so correct errors of lazzy webmasters
 */
char *robots::makeGoodToken(char *res) {
  assert(res != NULL);
  switch (res[0]) {
  case '*':
    // it's forbidden to have an * here,
    // for excluding everything, the real syntax is '/'
    return newString("*");
    break;
  case '/':  // good case
    return newString(res);
    break;
  default:
    {
      // add / before the first char
      // since headers and content are in the same buffer, 
      // res-1 is always in a legal area
      res--;
      char c = res[0];
      res[0] = '/';
      char *ans = newString(res);
      res[0] = c;
      return ans;
    }
    break;
  }
}


/*************************************
 * implementation of html
 *************************************/

#ifdef SPECIFICSAVE
#define newSpec()  if (state == SPECIFIC) openSpec()
#define pipeSpec() writeSpec()
#else
#define newSpec() ((void) 0)
#define pipeSpec() 0
#endif // SPECIFICSAVE

/** Constructor
 */
html::html (url *here, Connexion *conn) : file(conn) {
  newPars();
  this->here = here;
  base = here->giveBase();
  state = ANSWER;
  isInteresting = false;
  pages();
#ifdef SPECIFICSAVE
  fdsSpec = -1;
  mydir = -1;   // indicate no save
#endif // SPECIFICSAVE
}

/** Destructor
 */
html::~html () {
  delPars();
  delete here;
  delete base;
#ifdef SPECIFICSAVE
  if (fdsSpec != -1) {
    close(fdsSpec);
  }
#endif // SPECIFICSAVE
}

/** a string is arriving, treat it only up to the end of headers
 * return 0 usually, 1 if no more input and set errno accordingly
 */
int html::inputHeaders (int size) {
  pos += size;
  buffer[pos] = 0;
  char *posn;
  while (posParse < buffer + pos) {
    switch (state) {
    case ANSWER:
      posn = index (posParse, '\n');
      if (posn != NULL) {
        posParse = posn;
        if (parseCmdline ()) {
          return 1;
        }
        area = ++posParse;
      } else {
        return 0;
      }
      break;
    case HEADERS:
    case HEADERS30X:
      posn = index (posParse, '\n');
      if (posn != NULL) {
        posParse = posn;
        int tmp;
        if (state == HEADERS)
          tmp = parseHeader();
        else tmp = parseHeader30X();
        if (tmp) {
          return 1;
        }
        area = ++posParse;
      } else {
        return 0;
      }
      break;
    case SPECIFIC:
      return pipeSpec();
    default:
      return 0;
    }
  }
  return 0;
}

/** file download is complete, parse the file (headers already done)
 * return 0 usually, 1 if there was an error
 */
int html::endInput () {
  if (state <= HEADERS) {
    errno = earlyStop;
    return 1;
  }
  if (state == HEADERS30X) {
    errno = err40X;
    return 1;
  }
#ifdef NO_DUP
  if (!global::hDuplicate->testSet(posParse)) {
    errno = duplicate;
    return 1;
  }
#endif // NO_DUP
  buffer[pos] = 0;
  if (state == SPECIFIC) return 0;
  // now parse the html
  char *posn;
  while (posParse < buffer + pos) {
    switch (state) {
    case HTML:
      posn = index (posParse, '<');
      if (posn != NULL) {
        state = TAG;
        tagId = NONE;
        tagPos = 1;
        posParse = posn + 1;
      } else {
        return 0;
      }
      break;
    case COMMENT:
      if (*posParse == "-->"[tagPos]) {
        tagPos++;
        if (tagPos == 3) {
          state = HTML;
        }
      } else {
        tagPos = 0;
      }
      posParse++;
      break;
    case TAG:
      // That's a tag
      if (*posParse!='\n' && *posParse!=' '
          && *posParse!='\t' && *posParse!='\r') {
        // interesting char
        matchTag(lowerCase(*posParse));
      }
      posParse++;
      break;
    case TAGCONTENT:
      // Content of the tag
      {
        if (area==posParse &&
            (*posParse=='\"' || *posParse==' ')) {
          posParse++; area++;
        }
        char *endItem = area + maxUrlSize;
        if (endItem > buffer + pos) endItem = buffer + pos;
        while (posParse < endItem && *posParse!='\"' && *posParse!='\''
               && *posParse!='\n' && *posParse!=' ' && *posParse!='>'
               && *posParse!='\r' && *posParse!='?' && *posParse!='='
               && *posParse!='\t' && *posParse!='*') {
          if (*posParse == '\\') *posParse = '/';    // Bye Bye DOS !
          posParse++;
        }
        if (posParse == buffer + pos) {
          // This input only includes the beginning of the content
          return 0;
        } else if (posParse < endItem && *posParse!='?'
                   && *posParse!='=' && *posParse!='*') {
          // compute this url
          tagContent ();
        } else {
          // this url if too long or is a cgi : forget it
          state = HTML;
        }
        posParse++;
      }
      break;
    default:
      cerr << "switch not exhausive in input(char *c, int size) (file.cc)\n";
      break;
    }
  }
  return 0;
}

/** parse the answer code line */
int html::parseCmdline () {
  if (posParse - buffer >= 12) {
    switch (buffer[9]) {
    case '2':
      state = HEADERS;
      break;
    case '3':
      state = HEADERS30X;
      break;
    default:
      errno = err40X;
      return 1;
    }
  } else {
    errno = earlyStop;
    return 1;
  }
  return 0;
}

/** parse a line of header
 * @return 0 if OK, 1 if we don't want to read the file
 */
int html::parseHeader () {
  if (posParse - area < 2) {
	// end of http headers
#ifdef FOLLOW_LINKS
	if (global::isSpecific && isInteresting) {
	  state = SPECIFIC;
	} else {
	  state = HTML;
	}
#else // FOLLOW_LINKS
    state = SPECIFIC;
#endif // FOLLOW_LINKS
    contentStart = posParse + 1;
    *(posParse-1) = 0;
    newSpec();
  } else {
	if (startWithIgnoreCase("Content-Type: ", area)) {
	  // Let's read the type of this doc
	  if (!startWithIgnoreCase("text/html", area+14)) {
		if (global::isSpecific
			&& startWithIgnoreCase(global::contentType, area+14)) {
		  interestingSeen();
		  isInteresting = true;
		} else {
		  errno = badType;
          return 1;
		}
	  }
	}
#ifndef SPECIFICSAVE
    else if (startWithIgnoreCase("Content-Length: ", area)) {
      int len = 0;
      char *p = area+16;
      while (*p >= '0' && *p <= '9') {
        len = len*10 + *p -'0';
        p++;
      }
      if (len > maxPageSize) {
        errno = tooBig;
        return 1;
      }
    }
#endif // SPECIFICSAVE
  }
  return 0;
}

/** parse a line of header (ans 30X) => just look for location
 * @return 0 if OK, 1 if we don't want to read the file
 */
int html::parseHeader30X () {
  if (posParse - area < 2) {
	// end of http headers without location => err40X
    errno = err40X;
    return 1;
  } else {
	if (startWithIgnoreCase("Location: ", area)) {
      int i=10;
      while (area[i]!=' ' && area[i]!='\n' && area[i]!='\r'
             && area[i]!='?' && area[i]!='=') {
        i++;
      }
      if (area[i]!='?' && area[i]!='=') {
        area[i] = 0; // end of url
        // read the location (do not decrease depth)
#ifdef URL_TAGS
        url *nouv = new url(area+10, here->getDepth(), here->tag, base);
#else
        url *nouv = new url(area+10, here->getDepth(), base);
#endif // URL_TAGS
        manageUrl(nouv);
        // we do not need more headers
      }
      errno = err30X;
      return 1;
	}
  }
  return 0;
}

/** Try to understand this tag
 */
void html::matchTag (char c) {
  switch (tagId) {
  case NONE:
	//no tag yet
	switch (c) {
	case 'a':
	  tagId = AHREF;
	  break;
	case 'b':
	  tagId = BASE;
	  break;
    case 'f':
      tagId = FRAME;
      break;
    case '!':
      tagId = COMMENT;
      break;
	default:
	  state = HTML;
	}
	break;
  case AHREF:
	// ahref tag
	if ("ahref="[tagPos] == c) {
	  tagPos++;
	  if (tagPos == 6) {
        area = posParse+1;
		state = TAGCONTENT;
	  }
	} else {
	  state = HTML;
	}
	break;
  case FRAME:
	// framesrc tag
    if (c == '>') {
	  state = HTML;
	} else if ("framese"[tagPos] == c) {
      tagPos++;
      if (tagPos == 7) {
        // this is the frameset tag => forget
        state = HTML;
      }
    } else if (c == 'r') {
      tagPos = 2;
    } else if ("src="[tagPos] == c) {
      tagPos++;
      if (tagPos == 4) {
        tagId = AHREF;
        area = posParse+1;
		state = TAGCONTENT;
      }
    } else {
      tagPos = 0;
    }
	break;
  case BASE:
	// basehref tag
	if ("basehref="[tagPos] == c) {
	  tagPos++;
	  if (tagPos == 9) {
        area = posParse+1;
		state = TAGCONTENT;
	  }
	} else {
	  state = HTML;
	}
	break;
  case COMMENT:
	// basehref tag
	if ("!--"[tagPos] == c) {
	  tagPos++;
	  if (tagPos == 3) {
        tagPos = 0;
		state = COMMENT;
	  }
	} else {
	  state = HTML;
	}
	break;
  }
}

/* manage a new url : verify and send it */
void html::manageUrl (url *nouv) {
  if (nouv->isValid()) {
    nouv->setDepth(here->getHost());
    if (filter1(nouv->getHost(), nouv->getFile())) {
      // The extension is not stupid (gz, pdf...)
#ifdef LINKS_INFO
      links.addElement(nouv->giveUrl());
#endif // LINKS_INFO
      if (nouv->getDepth() >= 0) {
        check(nouv);
      } else {
        // We are too deep in this site
        delete nouv;
      }
    } else {
      // The extension is stupid
      delete nouv;
    }
  } else {
    // Bad url => delete it
    delete nouv;
  }
}

/** read the content of an interesting tag */
void html::tagContent () {
  char oldchar = *posParse;
  *posParse = 0;
  switch (tagId) {
  case AHREF:
    {
      // try to understand this new link
      uint dpth = here->getDepth();
#ifdef URL_TAGS
      url *nouv = new url(area, dpth-1, 0, base);
#else
      url *nouv = new url(area, dpth-1, base);
#endif // URL_TAGS
      manageUrl(nouv);
    }
    break;
  case BASE:
    // This page has a BASE HREF tag
    if (startWithIgnoreCase("http://", area)) {
      // http://host/a/b/c.html becomes host/a/b/
      uint end = posParse - area - 1;
      while (end >= 7 && area[end] != '/') {
        end--;
      }
      if (end > 7) {
        // the area looks good
        end++;
        char tmp = area[end];
        area[end] = 0;
        url *tmpbase = new url(area, 0);
        delete base;
        if (tmpbase->isValid()) {
          base = tmpbase;
        } else {
          delete tmpbase;
          base = NULL;
        }
        area[end] = tmp;
      }
    }
    break;
  }
  state = HTML;
  *posParse = oldchar;
}

///////////////////////////////////////////////////////////////
#ifdef SPECIFICSAVE

static int nbdir = -1;
static int nbfile = filesPerDir;
static char *fileName;
static uint endFileName;
static int indexFds = -1;
static char buf[maxUrlSize+30];

/** give the name of the file given dir and file number
 * this char * is static */
char *getSpecName(int nbdir, int nbfile) {
  sprintf(fileName+endFileName, "d%5i/f%5i%s",
          nbdir, nbfile, global::privilegedExt);
  for (int i=endFileName+1; fileName[i]==' '; i++)
    fileName[i] = '0';
  for (int i=endFileName+8; fileName[i]==' '; i++)
    fileName[i] = '0';
  return fileName;
}

char *getIndexName(int nbdir) {
  sprintf(fileName+endFileName, "d%5i/index", nbdir);
  for (int i=endFileName+1; fileName[i]==' '; i++)
    fileName[i] = '0';
  return fileName;
}

/** open file descriptor */
void html::openSpec () {
  nbfile++;
  if (nbfile >= filesPerDir) { // new dir
    nbdir++; nbfile = 0;
    // create the directory
    char *indexName = getIndexName(nbdir);
    fileName[endFileName+6] = 0;
    if (mkdir(fileName, S_IRWXU) != 0) perror("trouble while creating dir");
    fileName[endFileName+6] = '/';
    // open new index
    close(indexFds);
    indexFds = creat(indexName, S_IRWXU);
    if (indexFds < 0) {
      cerr << "cannot open file " << indexName << "\n";
      exit(1);
    }
  }
  mydir = nbdir;
  myfile = nbfile;
  // write index
  int s = sprintf(buf, "%4d  ", nbfile);
  s += here->writeUrl(buf+s);
  buf[s++] = '\n';
  ecrireBuff(indexFds, buf, s);
  // open new file
  char *fileName = getSpecName(nbdir, nbfile);
  fdsSpec = creat(fileName, S_IRWXU);
  if (fdsSpec < 0) {
    cerr << "cannot open file " << fileName << "\n";
    exit(1);
  }
  nbSpec = 0;
}

/** feed file descriptor */
bool html::writeSpec () {
  int nb = buffer + pos - posParse;
  ecrireBuff(fdsSpec, posParse, nb);
  pos = posParse - buffer;
  nbSpec += nb;
  if (nbSpec >= maxSpecSize) {
    errno = tooBig;
    return true;
  }
  return false;
}

/** in case of specific save */
void initSpecificSave () {
  mkdir(specDir, S_IRWXU);
  endFileName = strlen(specDir);
  fileName = new char[endFileName+20];
  strcpy(fileName, specDir);
  if (fileName[endFileName-1] != '/') fileName[endFileName++] = '/';
}
#endif // SPECIFICSAVE
