// Larbin
// Sebastien Ailleret
// 13-12-99 -> 19-11-01

#ifndef FILE_H
#define FILE_H

#include "types.h"
#include "utils/url.h"
#include "utils/string.h"
#include "utils/Vector.h"
#include "fetch/site.h"

struct Connexion;

class file {
 protected:
  // link to the buffer of our connexion
  char *buffer;
  // parsing position
  char *posParse;
 public:
  // Constructor
  file (Connexion *conn);
  // Destructor
  virtual ~file ();
  // current position in the buffer
  uint pos;
  // a string arrives from the server
  virtual int inputHeaders (int size) = 0; // just parse headers
  virtual int endInput () = 0;
  // Is it a robots.txt
  virtual bool isRobots ();
};

class html : public file {
 private:
  // Where are we
  url *here;
  // What is the current tag and how much of it did we already parse
  int tagId;
  int tagPos;
  // beginning of the current interesting area
  char *area;
  // begining of the real content (end of the headers + 1)
  char *contentStart;
  // base de l'URL
  url *base;
  /** parse the answer code line */
  int parseCmdline ();
  /* manage a new url : verify and send it */
  void manageUrl (url *nouv);
  /** parse a line of header (ans 30X) => just look for location
   * @return 0 if OK, 1 if we don't want to read the file
   */
  int parseHeader30X ();
  /** parse a line of header
   * @return 0 if OK, 1 if content type not understandable for us
   */
  int parseHeader ();
  /** Try to understand this tag
   */
  void matchTag (char c);
  /** read the content of an interesting tag
   */
  void tagContent ();

#ifdef LINKS_INFO
  /** links extracted from this page */
  Vector<char> links;
#endif // LINKS_INFO

#ifdef SPECIFICSAVE
  /** file descriptor number */
  int fdsSpec;
  /** number of bits written */
  int nbSpec;
  /** open file descriptor */
  void openSpec ();
  /** feed file descriptor,
   * return true and set errno = tooBig if necessary */
  bool writeSpec ();
#endif // SPECIFICSAVE

 public:
  // Constructor
  html (url *here, Connexion *conn);
  // Destructor
  ~html ();
  /** a string is arriving
   * return 0 usually, 1 if don't want any more input
   * in the latter case, errno is set to FetchError reason
   */
  int inputHeaders (int size); // just parse headers
  int endInput ();
  // State of our read : answer, headers, tag, html...
  int state;
  // Is it a robots.txt
  inline bool isRobots () { return false; }
  /** return the url of this file */
  inline url *getUrl () { return here; }
  /** Is this page interesting ? */
  bool isInteresting;
  /** return the content of the page */
  inline char *getPage () { return contentStart; }
  /** return the http headers */
  inline char *getHeaders () { return buffer; }

#ifdef LINKS_INFO
  /** return the links */
  inline Vector<char> *html::getLinks () { return &links; }
#endif // LINKS_INFO

#ifdef SPECIFICSAVE
  /** number of file and directory when saved on disk */
  int mydir;
  int myfile;
#endif // SPECIFICSAVE
};

class robots : public file {
 private:
  // did we parse the anser code
  bool answerCode;
  // test http headers
  bool parseHeaders ();
  // suppress commentary
  void uncomment ();
  // read the file
  void parseRobots ();
  /** Given a token, make it good (allocate)
   * try to understand it as a path, so correct errors of lazzy webmasters
   */
  char *makeGoodToken(char *tok);
 public:
  // Constructor
  robots (NamedSite *server, Connexion *conn);
  // Destructor
  ~robots ();
  // for which site is this robots
  NamedSite *server;
  // a string is arriving from the server
  int inputHeaders (int size); // just parse headers
  int endInput ();
  // parse the file (once everything has been read)
  void parse (bool isError);
};

// in case of specific save
#ifdef SPECIFICSAVE
/** various initialisations */
void initSpecificSave ();
/** give the name of the file given dir and file number
 * this char * is static */
char *getSpecName(int nbdir, int nbfile);
#else
#define initSpecificSave() ((void) 0)
#endif

#endif // FILE_H
