// Larbin
// Sebastien Ailleret
// 08-02-00 -> 22-07-01

#include <unistd.h>
#include <errno.h>
#include <iostream.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <adns.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <string.h>

#include "options.h"

#include "types.h"
#include "utils/Fifo.h"
#include "utils/debug.h"
#include "utils/text.h"
#include "utils/connexion.h"
#include "interf/output.h"
#include "fetch/site.h"


///////////////////////////////////////////////////////////
// class Interval
///////////////////////////////////////////////////////////

/** Constructor */
Interval::Interval (uint size) {
  this->size = size;
  pos = 0;
}

/** Destructor : never used */
Interval::~Interval () {
}


/////////////////////////////////////////////////////////
// functions used for the 2 types of sites
/////////////////////////////////////////////////////////

/** connect to this addr using connection conn 
 * return the state of the socket
 */
char getFds (Connexion *conn, sockaddr_in *addr) {
  // use proxy
  if (global::proxyAddr != NULL)
	return global::getProxyFds(conn);
  // no proxy
  int fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd < 0)
    return EMPTY;
  else
    global::verifMax(fd);
  conn->socket = fd;
  for (;;) {
    fcntl(fd, F_SETFL, O_NONBLOCK);
    if (connect(fd, (struct sockaddr*) addr,
                sizeof (struct sockaddr_in)) == 0) {
      // success
      return WRITE;
    } else if (errno == EINPROGRESS) {
      // would block
      return CONNECTING;
    } else {
      // error
      (void) close(fd);
      return EMPTY;
    }
  }
}


///////////////////////////////////////////////////////////
// class NamedSite
///////////////////////////////////////////////////////////

/** Constructor : initiate fields used by the program
 */
NamedSite::NamedSite () {
  name = newString("");
  forbidden = NULL;
  nburls = 0;
  tab = new Fifo<url>(fifoSiteSize);
  inFifo = false;
}

/** Destructor : This one is never used
 */
NamedSite::~NamedSite () {
  cerr << "Some named site is deleted (should not happen)\n";
  assert(false);
}

/** Put an url in the fifo
 * If there are too much, put it back in UrlsInternal (ie on disk)
 */
void NamedSite::putUrl (url *u) {
  if (nburls > maxUrlsBySite
	  && global::URLsInternal->getLength()*2 > ramUrls) {
	// Already enough Urls in memory for this Site
    // first check if it can already be forgotten
    if (!strcmp(name, u->getHost())
        && port == u->getPort()) {
      if (dnsState == errorDns) {
        nburls++;
        forgetUrl(u, noDNS);
        return;
      }
      if (dnsState == noConnDns) {
        nburls++;
        forgetUrl(u, noConnection);
        return;
      }
      if (!testRobots(u->getFile())) {
        nburls++;
        forgetUrl(u, forbiddenRobots);
        return;
      }
    }
    // else put it back in URLsInternal
    refUrl();
	global::URLsInternal->put(u);
	global::inter->getOne();
  } else {
    putPriorityUrl(u);
  }
}

/** Put an prioritarian url in the fifo
 * Up to now, it's very naive
 * because we have no memory of priority inside the url
 */
void NamedSite::putPriorityUrl (url *u) {
  nburls++;
  if (dnsState == waitDns
      || strcmp(name, u->getHost())
      || port != u->getPort()
      || global::now > lastUpdate + dnsValidTime) {
    tab->put(u);
    // Put Site in fifo if not yet in
    if (!inFifo) {
      inFifo = true;
      global::dnsSites->put(this);
    }
  } else switch (dnsState) {
  case doneDns:
    transfer(u);
    break;
  case errorDns:
    forgetUrl(u, noDNS);
    break;
  default: // noConnDns
    forgetUrl(u, noConnection);
  }
}

/** Init a new dns query
 */
void NamedSite::newQuery (uint *nbCalls) {
  // Update our stats
  newId();
  if (global::proxyAddr != NULL) {
    // we use a proxy, no need to get the sockaddr
    // give anything for going on
    siteSeen();
    siteDNS();
    // Get the robots.txt
    dnsOK();
  } else if (isdigit(name[0])) {
    // the name already in numbers-and-dots notation
	siteSeen();
	if (inet_aton(name, &addr.sin_addr)) {
	  // Yes, it is in numbers-and-dots notation
	  siteDNS();
	  // compute the new addr
	  addr.sin_family = AF_INET;
	  addr.sin_port = htons(port);
	  // Get the robots.txt
	  dnsOK();
	} else {
	  // No, it isn't : this site is a non sense
	  dnsErr();
	}
  } else {
    (*nbCalls)++;
    adns_query quer = NULL;
    crash("Submit an adns query");
    int res = adns_submit(global::ads, name,
                          (adns_rrtype) adns_r_addr,
                          (adns_queryflags) 0,
                          this, &quer);
    crash("End of submission");
    assert (res == 0);
  }
}

/** The dns query ended with success
 * assert there is a freeConn
 */
void NamedSite::dnsAns (adns_answer *ans) {
  siteSeen();
  if (ans->status != adns_s_ok) {
	// No addr inet
	dnsErr();
  } else {
	siteDNS();
	// compute the new addr
	memcpy (&addr,
			&ans->rrs.addr->addr.inet,
			sizeof (sockaddr_in));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	// Get the robots.txt
	dnsOK();
  }
}

/** we've got a good dns answer
 * get the robots.txt
 * assert there is a freeConn
 */
void NamedSite::dnsOK () {
  dnsState = doneDns;
  Connexion *conn = global::freeConns->get();
  char res = getFds(conn, &addr);
  if (res != EMPTY) {
    conn->timeout = global::now + timeoutPage;
    if (global::proxyAddr != NULL) {
      // use a proxy
      conn->request.addString("GET http://");
      conn->request.addString(name);
      char tmp[15];
      sprintf(tmp, ":%u", port);
      conn->request.addString(tmp);
      conn->request.addString("/robots.txt HTTP/1.0\r\nHost: ");
    } else {
      // direct connection
      conn->request.addString("GET /robots.txt HTTP/1.0\r\nHost: ");
    }
    conn->request.addString(name);
    conn->request.addString(global::headersRobots);
    conn->parser = new robots(this, conn);
    conn->pos = 0;
    conn->err = success;
    conn->state = res;
  } else {
    // Unable to get a socket
    global::freeConns->put(conn);
    dnsErr();
  }
}

/** Cannot get the inet addr
 */
void NamedSite::dnsErr () {
  dnsState = errorDns;
  int ss = tab->getLength();
  // scan the queue
  for (int i=0; i<ss; i++) {
    url *u = tab->get();
    if (!strcmp(name, u->getHost())
        && port == u->getPort()) {
      forgetUrl(u, noDNS);
    } else {
      tab->put(u);
    }
  }
  // where should now lie this site
  if (tab->isEmpty()) {
    inFifo = false;
  } else {
    global::dnsSites->put(this);
  }
}

/** test if a file can be fetched thanks to the robots.txt */
bool NamedSite::testRobots(char *file) {
  if (forbidden == NULL) {
    return true;
  } else {
    int i=0;
    while ((*forbidden)[i] != NULL) {
      if (startWith((*forbidden)[i], file)) {
        return false;
      }
      i++;
    }
    return true;
  }
}

/** Delete the old identity of the site */
void NamedSite::newId () {
  // Change the identity of this site
#ifndef NDEBUG
  if (name[0] == 0) {
    addsite();
  }
#endif // NDEBUG
  delete [] name;
  url *u = tab->read();
  name = newString(u->getHost());
  port = u->getPort();
  lastUpdate = global::now;
  if (forbidden != NULL) {
    delete forbidden;
    forbidden = NULL;
  }
  dnsState = waitDns;
}

/** we got the robots.txt, transfer what must be in IPSites */
void NamedSite::robotsResult (FetchError res) {
  if (res == noConnection) dnsState = noConnDns;
  int ss = tab->getLength();
  // scan the queue
  for (int i=0; i<ss; i++) {
    url *u = tab->get();
    if (!strcmp(name, u->getHost())
        && port == u->getPort()) {
      if (res == noConnection) {
        forgetUrl(u, noConnection);
      } else {
        transfer(u);
      }
    } else {
      tab->put(u);
    }
  }
  // where should now lie this site
  if (tab->isEmpty()) {
    inFifo = false;
  } else {
    global::dnsSites->put(this);
  }  
}

void NamedSite::transfer (url *u) {
  if (testRobots(u->getFile())) {
    uint h=0;
    if (global::proxyAddr == NULL) {
      memcpy (&u->addr, &addr, sizeof (sockaddr_in)); 
      char *s = (char *) &addr;
      for (uint i=0; i<sizeof(sockaddr_in); i++) {
        h = h*31 + s[i];
      }
      h %= IPSiteListSize;
    }
    global::IPSiteList[h].putUrl(u);
  } else {
    forgetUrl(u, forbiddenRobots);
  }
}

void NamedSite::forgetUrl (url *u, FetchError reason) {
  urls();
  fetchFail(u, reason);
  answers(reason);
  nburls--;
  delete u;
  global::inter->getOne();
}

///////////////////////////////////////////////////////////
// class IPSite
///////////////////////////////////////////////////////////

/** Constructor : initiate fields used by the program
 */
IPSite::IPSite () {
  tab = new Fifo<url>(fifoSiteSize);
  lastAccess = 0;
  inFifo = false;
}

/** Destructor : This one is never used
 */
IPSite::~IPSite () {
  cerr << "Some ip site is deleted (should not happen)\n";
  assert(false);
}

/** Put an prioritarian url in the fifo
 * Up to now, it's very naive
 * because we have no memory of priority inside the url
 */
void IPSite::putUrl (url *u) {
  // All right, put this url inside at the end of the queue
  tab->put(u);
  // Put Site in fifo if not yet in
  if (!inFifo) {
#ifndef NDEBUG
    if (lastAccess == 0) addipsite();
#endif // NDEBUG
    inFifo = true;
    if (lastAccess + global::waitDuration <= global::now
        && global::freeConns->isNonEmpty()) {
      fetch();
    } else {
      global::okSites->put(this);
    }
  }
}

/** Get an url from the fifo and do some stats
 */
inline url *IPSite::getUrl () {
  url *u = tab->get();
  urls();
  global::namedSiteList[u->hostHashCode()].nburls--;
  global::inter->getOne();
#ifndef NOSTATS
  int len = strlen(u->getFile());
  if (global::isSpecific
      && global::privilegedExt != NULL
      && endWithIgnoreCase(global::privilegedExt, u->getFile(), len)) {
    extensionTreated();
  }
#endif
  return u;
}

/** fetch the first page in the fifo okSites
 * there must be at least one element in freeConns !!!
 * return expected time for next call (0 means now is OK)
 * This function always put the IPSite in fifo before returning
 *   (or set inFifo to false if empty)
 */
int IPSite::fetch () {
  if (tab->isEmpty()) {
	// no more url to read
	// This is possible because this function can be called recursively
	inFifo = false;
    return 0;
  } else {
    int next_call = lastAccess + global::waitDuration;
    if (next_call > global::now) {
      global::okSites->rePut(this);
      return next_call;
    } else {
      Connexion *conn = global::freeConns->get();
      url *u = getUrl();
      // We're allowed to fetch this one
      // open the socket
      char res = getFds(conn, &(u->addr));
      if (res != EMPTY) {
        lastAccess = global::now;
        conn->timeout = lastAccess + timeoutPage;
        conn->request.addString("GET ");
        if (global::proxyAddr != NULL) {
          char *tmp = u->giveUrl();
          conn->request.addString(tmp);
          delete [] tmp;
        } else {
          conn->request.addString(u->getFile());
        }
        conn->request.addString(" HTTP/1.0\r\nHost: ");
        conn->request.addString(u->getHost());
        conn->request.addString(global::headers);
        conn->parser = new html (u, conn);
        conn->pos = 0;
        conn->err = success;
        conn->state = res;
        if (tab->isEmpty()) {
          inFifo = false;
        } else {
          global::okSites->put(this);
        }
        return 0;
      } else {
        // Unable to connect
        fetchFail(u, noConnection);
        answers(noConnection);
        delete u;
        global::freeConns->put(conn);
        return fetch();
      }    
    }
  }
}
