// Larbin
// Sebastien Ailleret
// 08-02-00 -> 01-05-01

// This is the new structure of a site
// It includes a fifo of waiting urls

#ifndef SITE_H
#define SITE_H

#include <time.h>

#include "types.h"
#include "global.h"
#include "utils/Fifo.h"
#include "utils/url.h"

enum DnsState {
  waitDns,
  doneDns,
  errorDns,
  noConnDns
};

/** This class is intended to make sure the sum of the
 * sizes of the fifo included in the different sites
 * are not too big
 */
class Interval {
  private:
  /** Position in the interval */
  uint pos;
  /** Size of the interval */
  uint size;
 public:
  /** Constructor */
  Interval (uint size);
  /** Destructor */
  ~Interval ();
  /** How many urls can we put
   * answer 0 if no urls can be put
   */
  inline uint putAll () { int res=size-pos; pos=size; return res; }
  /** Warn an url has been retrieved */
  inline void getOne () { pos--; }
  /** only for debugging, handle with care */
  inline uint getPos () { return pos; }
};

class NamedSite {
 public:
  /* name of the site */
  char *name;
  /* port of the site */
  uint port;
  /** internet addr of this server */
  sockaddr_in addr;
  DnsState dnsState;
  /* Date of dns call and robots.txt fetch */
  time_t lastUpdate;
  /** Urls waiting for being fetched */
  Fifo<url> *tab;
  /** Is this Site in a dnsSites */
  bool inFifo;
  /** we've got a good dns answer
   * get the robots.txt
   */
  void dnsOK ();
  /** Cannot get the inet addr */
  void dnsErr ();
  /** Delete the old identity of the site */
  void newId ();
  /** test if a file can be fetched thanks to the robots.txt */
  bool testRobots(char *file);
  /** put this url in its IPSite */
  void NamedSite::transfer (url *u);
  /** forget this url for this reason */
  void forgetUrl (url *u, FetchError reason);
 public:
  /** Constructor */
  NamedSite ();
  /** Destructor : never used */
  ~NamedSite ();
  /* forbidden paths : given by robots.txt */
  Vector<char> *forbidden;
  /* numbers of urls in ram for this site */
  uint nburls;
  /** Put an url in the fifo
   * If there are too much, put it back in UrlsInternal
   */
  void putUrl (url *u);
  /** Put an prioritarian url in the fifo */
  void putPriorityUrl (url *u);
  /** Init a new dns query */
  void newQuery (uint *nbCalls);
  /** The dns query ended with success */
  void dnsAns (adns_answer *ans);
  /** we got the robots.txt, transfer what must be in IPSites */
  void robotsResult (FetchError res);
};

class IPSite {
 private:
  /* date of last access : avoid rapid fire */
  time_t lastAccess;
  /** Urls waiting for being fetched */
  Fifo<url> *tab;
  /** Is this Site in a okSites (eg have something to fetch) */
  bool inFifo;
  /** Get an url from the fifo
   * resize tab if too big
   */
  url *getUrl ();
 public:
  /** Constructor */
  IPSite ();
  /** Destructor : never used */
  ~IPSite ();
  /** Put an url in the fifo */
  void putUrl (url *u);
  /** fetch the fist page in the fifo okSites
   * expects at least one element in freeConns
   * return expected time for next call (0 means now)
   */
  int fetch ();
};

#endif // SITE_H
