// Larbin
// Sebastien Ailleret
// 29-11-99 -> 19-11-01

#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <fcntl.h>
#include <iostream.h>
#include <string.h>
#include <adns.h>
#include <netdb.h>
#include <sys/socket.h>
#include <string.h>
#include <signal.h>

#include "options.h"

#include "types.h"
#include "global.h"
#include "utils/text.h"
#include "utils/Fifo.h"
#include "utils/debug.h"
#include "fetch/site.h"
#include "interf/output.h"


///////////////////////////////////////////////////////////
// Struct global
///////////////////////////////////////////////////////////

// define all the static variables
time_t global::now;
hashTable *global::seen;
#ifdef NO_DUP
hashDup *global::hDuplicate;
#endif // NO_DUP
SyncFifo<url> *global::URLsInput;
PersistentFifo *global::URLsInternal;
IPSite *global::IPSiteList;
NamedSite *global::namedSiteList;
Fifo<IPSite> *global::okSites;
Fifo<NamedSite> *global::dnsSites;
Connexion *global::connexions;
adns_state global::ads;
ConstantSizedFifo<Connexion> *global::freeConns;
#ifdef THREAD_OUTPUT
ConstantSizedFifo<Connexion> *global::userConns;
#endif
Interval *global::inter;
uint global::depthInSite;
time_t global::waitDuration;
char *global::userAgent;
char *global::sender;
char *global::headers;
char *global::headersRobots;
sockaddr_in *global::proxyAddr;
bool global::isSpecific;
char *global::contentType;
char *global::privilegedExt;
Vector<char> *global::domains;
Vector<char> global::forbExt;
char *global::firstUrl;
uint global::nb_conn;
uint global::dnsConn;
unsigned short int global::httpPort;
unsigned short int global::inputPort;
int global::inputFds;
int global::nbInput;
Input *global::inputConns[maxInput];
struct pollfd *global::pollfds;
uint global::posPoll;
uint global::sizePoll;
short *global::ansPoll;
int global::maxFds;
#ifdef MAXBANDWIDTH
long int global::remainBand = MAXBANDWIDTH;
#endif // MAXBANDWIDTH

/** Constructor : initialize almost everything
 * Everything is read from the config file (larbin.conf by default)
 */
global::global (int argc, char *argv[]) {
  char *configFile = "larbin.conf";
#ifdef RELOAD
  bool reload = true;
#else
  bool reload = false;
#endif
  now = time(NULL);
  // verification of arguments
  int pos = 1;
  while (pos < argc) {
	if (!strcmp(argv[pos], "-c") && argc > pos+1) {
	  configFile = argv[pos+1];
	  pos += 2;
	} else if (!strcmp(argv[pos], "-scratch")) {
	  reload = false;
	  pos++;
	} else {
	  break;
	}
  }
  if (pos != argc) {
	cerr << "usage : " << argv[0];
	cerr << " [-c configFile] [-scratch]\n";
	exit(1);
  }

  // Standard values
  waitDuration = 60;
  depthInSite = 5;
  userAgent = "larbin";
  sender = "larbin@unspecified.mail";
  firstUrl = NULL;
  nb_conn = 20;
  dnsConn = 3;
  httpPort = 8081;
  inputPort = 0;  // by default, no input available
  proxyAddr = NULL;
  isSpecific = false;
  domains = NULL;
  // Read the configuration file
  crash("Read the configuration file");
  parseFile(configFile);
  // Initialize everything
  crash("Create global values");
  // Headers
  String strtmp;
  strtmp.addString("\r\nUser-Agent: ");
  strtmp.addString(userAgent);
  strtmp.addString(" ");
  strtmp.addString(sender);
  if (isSpecific) {
    strtmp.addString("\r\nAccept: text/html, ");
    strtmp.addString(contentType);
    strtmp.addString("\r\n\r\n");
  } else {
    strtmp.addString("\r\nAccept: text/html\r\n\r\n");
  }
  headers = strtmp.giveString();
  // Headers robots.txt
  String strtmp2;
  strtmp2.addString("\r\nUser-Agent: ");
  strtmp2.addString(userAgent);
  strtmp2.addString(" (");
  strtmp2.addString(sender);
  strtmp2.addString(")\r\n\r\n");
  headersRobots = strtmp2.giveString();
  // FIFOs
  URLsInternal = new PersistentFifo(reload);
  URLsInput = new SyncFifo<url>;
  inter = new Interval(ramUrls);
  namedSiteList = new NamedSite[namedSiteListSize];
  IPSiteList = new IPSite[IPSiteListSize];
  okSites = new Fifo<IPSite>(1000);
  dnsSites = new Fifo<NamedSite>(500);
  seen = new hashTable(!reload);
#ifdef NO_DUP
  hDuplicate = new hashDup(dupSize, dupFile, !reload);
#endif // NO_DUP
#ifdef THREAD_OUTPUT
  userConns = new ConstantSizedFifo<Connexion>(nb_conn);
#endif
  freeConns = new ConstantSizedFifo<Connexion>(nb_conn);
  connexions = new Connexion [nb_conn];
  for (uint i=0; i<nb_conn; i++) {
	freeConns->put(connexions+i);
  }
  // init poll structures
  sizePoll = nb_conn + maxInput;
  pollfds = new struct pollfd[sizePoll];
  posPoll = 0;
  maxFds = sizePoll;
  ansPoll = new short[maxFds];
  // init non blocking dns calls
  crash("Start adns");
  adns_initflags flags =
	adns_initflags (adns_if_nosigpipe | adns_if_noerrprint);
  adns_init(&ads, flags, NULL);
  // input
  if (inputPort != 0) {
    int allowReuse = 1;
    struct sockaddr_in addr;
    bzero (&addr, sizeof(addr));
    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_family = AF_INET;
    addr.sin_port = htons(inputPort);
    if ((inputFds = socket(AF_INET, SOCK_STREAM, 0)) == -1
        || setsockopt(inputFds, SOL_SOCKET, SO_REUSEADDR,
                      (char*)&allowReuse, sizeof(allowReuse))
        || bind(inputFds, (struct sockaddr *) &addr, sizeof(addr)) != 0
        || listen(inputFds, 4) != 0) {
      cerr << "Unable to get the socket\n";
      exit(1);
    }
    fcntl(inputFds, F_SETFL, O_NONBLOCK);
    for (int i=0; i<maxInput; i++) {
      inputConns[i] = new Input;
    }
    nbInput = 0;
  } else {
    nbInput = -1;
  }
  // prepare save strustures
  initSave();
  initSpecificSave();
  // let's ignore SIGPIPE
  static struct sigaction sn, so;
  sigemptyset(&sn.sa_mask);
  sn.sa_flags = SA_RESTART;
  sn.sa_handler = SIG_IGN;
  if (sigaction(SIGPIPE, &sn, &so)) {
    cerr << "Unable to disable SIGPIPE\n";
  }
}

/** Destructor : never used because the program should never end !
 */
global::~global () {
  assert(false);
}

/** parse configuration file */
void global::parseFile (char *file) {
  int fds = open(file, O_RDONLY);
  if (fds < 0) {
	cerr << "cannot open config file\n";
	exit(1);
  }
  char *tmp = readfile(fds);
  close(fds);
  // suppress commentary
  bool eff = false;
  for (int i=0; tmp[i] != 0; i++) {
	switch (tmp[i]) {
	case '\n': eff = false; break;
	case '#': eff = true; // no break !!!
	default: if (eff) tmp[i] = ' ';
	}
  }
  String content;
  content.addString(tmp);
  delete [] tmp;
  uint pos = 0;
  char *tok = nextTokenStr(content, &pos);
  while (tok != NULL) {
	if (!strcasecmp(tok, "UserAgent")) {
	  delete [] tok;
	  userAgent = nextTokenStr(content, &pos);
	} else if (!strcasecmp(tok, "From")) {
	  delete [] tok;
	  sender = nextTokenStr(content, &pos);
	} else if (!strcasecmp(tok, "startUrl")) {
	  delete [] tok;
	  firstUrl = nextTokenStr(content, &pos);
	} else if (!strcasecmp(tok, "waitduration")) {
	  delete [] tok;
	  tok = nextTokenStr(content, &pos);
	  waitDuration = atoi(tok);
	  delete [] tok;
	} else if (!strcasecmp(tok, "proxy")) {
	  delete [] tok;
	  // host name and dns call
	  tok = nextTokenStr(content, &pos);
	  struct hostent* hp;
	  proxyAddr = new sockaddr_in;
	  bzero((char *) proxyAddr, sizeof (struct sockaddr_in));
	  if ((hp = gethostbyname(tok)) == NULL) {
		endhostent();
		cerr << "Unable to find proxy ip address\n";
		exit(1);
	  } else {
		proxyAddr->sin_family = hp->h_addrtype;
		memcpy ((char*) &proxyAddr->sin_addr, hp->h_addr, hp->h_length);
	  }
	  endhostent();
	  delete [] tok;
	  // port number
	  tok = nextTokenStr(content, &pos);
	  proxyAddr->sin_port = htons(atoi(tok));
	  delete [] tok;
	} else if (!strcasecmp(tok, "pagesConnexions")) {
	  delete [] tok;
	  tok = nextTokenStr(content, &pos);
	  nb_conn = atoi(tok);
	  delete [] tok;
	} else if (!strcasecmp(tok, "dnsConnexions")) {
	  delete [] tok;
	  tok = nextTokenStr(content, &pos);
	  dnsConn = atoi(tok);
	  delete [] tok;
	} else if (!strcasecmp(tok, "httpPort")) {
	  delete [] tok;
	  tok = nextTokenStr(content, &pos);
	  httpPort = atoi(tok);
	  delete [] tok;
	} else if (!strcasecmp(tok, "inputPort")) {
	  delete [] tok;
	  tok = nextTokenStr(content, &pos);
	  inputPort = atoi(tok);
	  delete [] tok;
	} else if (!strcasecmp(tok, "depthInSite")) {
	  delete [] tok;
	  tok = nextTokenStr(content, &pos);
	  depthInSite = atoi(tok);
	  delete [] tok;
	} else if (!strcasecmp(tok, "specificSearch")) {
	  delete [] tok;
	  if (isSpecific) {
		cerr << "Bad configuration file : Two specificSearch fields\n";
		exit(1);
	  }
	  isSpecific = true;
	  contentType = nextTokenStr(content, &pos);
	  privilegedExt = nextTokenStr(content, &pos);
	  if (privilegedExt != NULL) {
        if (!strcasecmp(privilegedExt, "end")) {
          delete [] privilegedExt;
          privilegedExt = NULL;
        } else {
          int l = strlen(privilegedExt);
          int i;
          for (i=0; i<l; i++) {
            privilegedExt[i] = privilegedExt[i] | 32;
          }
          tok = nextTokenStr(content, &pos);
          if (tok == NULL || strcasecmp(tok, "end")) {
            cerr << "Bad configuration file : no end to specificSearch\n";
            exit(1);
          }
          delete [] tok;
        }
	  } else {
        cerr << "Bad configuration file : no end to specificSearch\n";
	  }
	} else if (!strcasecmp(tok, "limitToDomain")) {
	  delete [] tok;
	  manageDomain(content, &pos);
	} else if (!strcasecmp(tok, "forbiddenExtensions")) {
	  delete [] tok;
	  manageExt(content, &pos);
	} else {
	  cerr << "bad configuration file : " << tok << "\n";
	  exit(1);
	}
	tok = nextTokenStr(content, &pos);
  }
}

/** read the domain limit */
void global::manageDomain (String &content, uint *pos) {
  char *tok = nextTokenStr(content, pos);
  if (domains == NULL) {
	domains = new Vector<char>;
  }
  while (tok != NULL && strcasecmp(tok, "end")) {
	domains->addElement(tok);
	tok = nextTokenStr(content, pos);
  }
  if (tok == NULL) {
	cerr << "Bad configuration file : no end to limitToDomain\n";
	exit(1);
  } else {
	delete [] tok;
  }
}

/** read the forbidden extensions */
void global::manageExt (String &content, uint *pos) {
  char *tok = nextTokenStr(content, pos);
  while (tok != NULL && strcasecmp(tok, "end")) {
    int l = strlen(tok);
    int i;
    for (i=0; i<l; i++) {
      tok[i] = tok[i] | 32;
    }
	forbExt.addElement(tok);
	tok = nextTokenStr(content, pos);
  }
  if (tok == NULL) {
	cerr << "Bad configuration file : no end to forbiddenExtensions\n";
	exit(1);
  } else {
	delete [] tok;
  }
}

/** connect to this server using connection conn 
 * return the state of the socket
 */
char global::getProxyFds (Connexion *conn) {
  assert (proxyAddr != NULL);
  int fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd < 0)
    return EMPTY;
  else
    verifMax(fd);
  conn->socket = fd;
  for (;;) {
	fcntl(fd, F_SETFL, O_NONBLOCK);
    if (connect(fd, (struct sockaddr*) proxyAddr,
                sizeof(struct sockaddr_in)) == 0) {
      // success
      return WRITE;
    } else if (errno == EINPROGRESS) {
		// would block
		return CONNECTING;
    } else {
		// error
		(void) close(fd);
		return EMPTY;
    }
  }
}

/** set this fds for next poll */
void global::setPoll (int fd, short event) {
  pollfds[posPoll].fd = fd;
  pollfds[posPoll].events = event;
  posPoll++;
  assert (posPoll < sizePoll);
}

/** make sure the max fds has not been reached */
void global::verifMax (int fd) {
  if (fd >= maxFds) {
    int n = 2 * maxFds;
    if (fd >= n) {
      n = fd + maxFds;
    }
    short *tmp = new short[n];
    for (int i=0; i<maxFds; i++) {
      tmp[i] = ansPoll[i];
    }
    for (int i=maxFds; i<n; i++) {
      tmp[i] = 0;
    }
    delete (ansPoll);
    maxFds = n;
    ansPoll = tmp;
  }
}

///////////////////////////////////////////////////////////
// Struct Connexion
///////////////////////////////////////////////////////////

/** put Connection in a coherent state
 */
Connexion::Connexion () {
  state = EMPTY;
  parser = NULL;
}

/** Destructor : never used : we recycle !!!
 */
Connexion::~Connexion () {
  assert(false);
}

/** Recycle a connexion
 */
void Connexion::recycle () {
  delete parser;
  request.recycle();
}
