// Larbin
// Sebastien Ailleret
// 18-11-99 -> 22-07-01

/* This class contains all global variables */

#ifndef GLOBAL_H
#define GLOBAL_H

#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>

#include <adns.h>

#include "fetch/file.h"
#include "fetch/hashTable.h"
#include "utils/hashDup.h"
#include "utils/url.h"
#include "utils/Vector.h"
#include "utils/string.h"
#include "utils/ConstantSizedFifo.h"
#include "utils/PersistentFifo.h"
#include "utils/ConstantSizedFifo.h"
#include "utils/SyncFifo.h"
#include "utils/Fifo.h"


// define for the state of a connection
#define EMPTY 0
#define CONNECTING 1
#define WRITE 2
#define OPEN 3

class NamedSite;
class IPSite;
class Interval;

/** This represent a connection : we have a fixed number of them
 * fetchOpen links them with servers
 * fetchPipe reads those which are linked
 */
struct Connexion {
  char state;      // what about this socket : EMPTY, CONNECTING, WRITE, OPEN
  int pos;         // What part of the request has been sent
  FetchError err;  // How did the fetch terminates
  int socket;      // number of the fds
  time_t timeout;  // timeout for this connexion
  String request;  // what is the http request
  file *parser;    // parser for the connexion (a robots.txt or an html file)
  char buffer[maxPageSize];
  /** Constructor */
  Connexion ();
  /** Dectructor : it is never used since we reuse connections */
  ~Connexion ();
  /** Recycle a connexion
   */
  void recycle ();
};

/* input connexion */
struct Input {
  int fds;
  uint pos;
  uint end_pos;
  uint end_posp;
  int priority;
  uint depth;
  uint test;
  char buffer[BUF_SIZE];
};

class global {
 public:
  /** Constructor : see global.cc for details */
  global (int argc, char * argv[]);
  /** Destructor : never used */
  ~global ();
  /** current time : avoid to many calls to time(NULL) */
  static time_t now;
  /** List of pages allready seen (one bit per page) */
  static hashTable *seen;
#ifdef NO_DUP
  /** Hashtable for suppressing duplicates */
  static hashDup *hDuplicate;
#endif // NO_DUP
  /** URLs for the sequencer : each one has a different priority */
  static SyncFifo<url> *URLsInput;
  /** This one has the lowest priority */
  static PersistentFifo *URLsInternal;
  /** hashtables of the site we accessed (cache) */
  static NamedSite *namedSiteList;
  static IPSite *IPSiteList;
  /** Sites which have at least one url to fetch */
  static Fifo<IPSite> *okSites;
  /** Sites which have at least one url to fetch
   * but need a dns call
   */
  static Fifo<NamedSite> *dnsSites;
  /** Informations for the fetch
   * This array contain all the connections (empty or not)
   */
  static Connexion *connexions;
  /** Internal state of adns */
  static adns_state ads;
  /** free connection for fetchOpen : connections with state==EMPTY */
  static ConstantSizedFifo<Connexion> *freeConns;
#ifdef THREAD_OUTPUT
  /** free connection for fetchOpen : connections waiting for end user */
  static ConstantSizedFifo<Connexion> *userConns;
#endif
  /** Sum of the sizes of a fifo in Sites */
  static Interval *inter;
  /** How deep should we go inside a site */
  static uint depthInSite;
  /** how many seconds should we wait beetween 2 calls at the same server 
   * 0 if you are only on a personnal server, >=30 otherwise
   */
  static time_t waitDuration;
  /** Name of the bot */
  static char *userAgent;
  /** Name of the man who lauch the bot */
  static char *sender;
  /** http headers to send with requests 
   * sends name of the robots, from field...
   */
  static char *headers;
  static char *headersRobots;  // used when asking a robots.txt
  /* internet address of the proxy (if any) */
  static sockaddr_in *proxyAddr;
  /** connect to this server through a proxy using connection conn 
   * return >0 in case of success (connecting or connected), 0 otherwise
   */
  static char getProxyFds (Connexion *conn);
  /** Are we doing a specific search
   */
  static bool isSpecific;
  /** What is the content-type we look for
   */
  static char *contentType;
  /** Is there a privileged extension
   */
  static char *privilegedExt;
  /** Limit to domain */
  static Vector<char> *domains;
  /** forbidden extensions
   * extensions which are allways to avoid : .ps, .pdf...
   */
  static Vector<char> forbExt;
  /** First URL to fetch : initiate the search */
  static char *firstUrl;
  /** number of parallel connexions
   * your kernel must support a little more than nb_conn file descriptor
   * select should be able to handle as many fds
   */
  static uint nb_conn;
  /** number of parallel dns calls */
  static uint dnsConn;
  /** port on which is launched the http statistic webserver */
  static unsigned short int httpPort;
  /////////  INPUT  ///////////////////////////////
  /** port on which input wait for queries */
  static unsigned short int inputPort;
  /** socket used for input */
  static int inputFds;
  /** number of opened input connections */
  static int nbInput;
  /** array for the opened input connections */
  static Input *inputConns[maxInput];
  /** parse configuration file */
  static void parseFile (char *file);
  /** read the domain limit */
  static void manageDomain (String &content, uint *pos);
  /** read the forbidden extensions */
  static void manageExt (String &content, uint *pos);
  /////////// POLL ///////////////////////////////////
  /** array used by poll */
  static struct pollfd *pollfds;
  /** pos of the max used field in pollfds */
  static uint posPoll;
  /** size of pollfds */
  static uint sizePoll;
  /** array used for dealing with answers */
  static short *ansPoll;
  /** number of the biggest file descriptor */
  static int maxFds;
  /** set a fds with an event for next poll */
  static void setPoll (int fd, short event);
  /** make sure the new socket is not too big for ansPoll */
  static void verifMax (int fd);
#ifdef MAXBANDWIDTH
  /** number of bits still allowed during this second */
  static long int remainBand;
#endif // MAXBANDWIDTH
};

#endif // GLOBAL_H
