// Larbin
// Sebastien Ailleret
// 03-02-00 -> 22-07-01

#include <unistd.h>
#include <iostream.h>
#include <string.h>

#include "options.h"

#include "types.h"
#include "global.h"
#include "utils/text.h"
#include "utils/debug.h"

#define INIT -1
#define END -2

static int nbp=0;
static int nbr=0;

static bool readMore (Input *in);
static char *readline (Input *in);

int input () {
  if (global::nbInput < 0) {
    // Input is disabled
    return -1;
  }
  int n = -1;
  if (global::nbInput < maxInput-1
      && global::ansPoll[global::inputFds]) {
    // test if there is a new connection
    struct sockaddr_in addr;
    int fdc;
    uint len = sizeof(addr);
    fdc = accept(global::inputFds, (struct sockaddr *) &addr, &len);
    if (fdc != -1) {
      fcntl(fdc, F_SETFL, O_NONBLOCK);
      global::inputConns[global::nbInput]->fds = fdc;
      global::inputConns[global::nbInput]->pos = 0;
      global::inputConns[global::nbInput]->end_pos = 0;
      global::inputConns[global::nbInput]->end_posp = 0;
      global::inputConns[global::nbInput]->priority = INIT;
      global::nbInput++;
#ifdef URL_TAGS
      ecrire(fdc, "Welcome to larbin input system !\nYour first line should look like \"priority:0 depth:5 test:1\"\nThe following should contain one id and one url separed by one space per line\n\"137 http://pauillac.inria.fr/~ailleret/prog/larbin/\" for instance\n\n");
#else
      ecrire(fdc, "Welcome to larbin input system !\nYour first line should look like \"priority:0 depth:5 test:1\"\nThe following should contain one url per line (http://pauillac.inria.fr/~ailleret/prog/larbin/ for instance)\n\n");
#endif
    }
  }
  if (global::nbInput < maxInput-1) {
    n = global::inputFds;
    global::setPoll(global::inputFds, POLLIN);
  }
  // read open sockets
  int i=0; 
  while (i<global::nbInput) {
    Input *in = global::inputConns[i];
    if (global::ansPoll[in->fds] && readMore(in)) {
      char *line = readline(in);
      while (line != NULL) {
        nbp++;
        if (in->priority == INIT) {
          // first line
          if (sscanf(line, "priority:%d depth:%u test:%u",
                     &in->priority, &in->depth, &in->test) == 3) {
            line = readline(in);
          } else {
            ecrire(in->fds, "Incorrect input\n");
            line = NULL;
            in->priority = END;
          }
        } else {
          // this is an url
#ifdef URL_TAGS
          int i=0;
          uint tag = 0;
          while (line[i] >= '0' && line[i] <= '9') {
            tag = 10*tag + line[i] - '0';
            i++;
          }
          url *u = new url(line+i+1, in->depth, tag);
#else
          url *u = new url(line, in->depth);
#endif // URL_TAGS
          if (u->isValid()) {
            if (in->test) {
              if (global::seen->testSet(u)) {
                hashUrls();   // stats
                if (in->priority) {
                  global::URLsInput->put(u);
                } else {
                  global::URLsInternal->put(u);
                }
              } else {
                delete u;
              }
            } else {
              hashUrls();   // stats
              global::seen->set(u);
              if (in->priority) {
                global::URLsInput->put(u);
              } else {
                global::URLsInternal->put(u);
              }
            }
          } else {
            delete u;
          }
          line = readline(in);            
        }
      }
    }
    if (in->priority == END) {
      // forget this connection
      ecrire(in->fds, "Bye bye...\n");
      close(in->fds);
      global::nbInput--;
      Input *tmp = global::inputConns[i];
      global::inputConns[i] = global::inputConns[global::nbInput];
      global::inputConns[global::nbInput] = tmp;
    } else { // go to next connection
      if (in->fds > n) n = in->fds;
      global::setPoll(in->fds, POLLIN);
      i++;
    }
  }
  return n;
}

static bool readMore (Input *in) {
  assert (in->end_posp == in->end_pos);
  if (in->end_posp - in->pos > maxUrlSize+100) {
    // error -> stop connection
    ecrire(in->fds, "Url submitted too long\n");
    in->priority = END;
    return false;
  }
  if (2 * in->pos > BUF_SIZE) {
    in->end_pos -= in->pos;
    in->end_posp = in->end_pos;
    memmove(in->buffer, in->buffer+in->pos, in->end_pos);
    in->pos = 0;
  }
  int nb = read(in->fds, in->buffer+in->end_pos, BUF_SIZE-in->end_pos);
  if (nb == -1 && errno == EAGAIN) {
    return false;
  } else if (nb <= 0) {
    in->priority = END;
    return false;
  } else {
    nbr += nb;
    in->end_pos += nb;
    return true;
  }
}

/* no allocation */
static char *readline (Input *in) {
  while (in->end_posp < in->end_pos && in->buffer[in->end_posp] != '\n') {
    in->end_posp++;
  }
  if (in->end_posp == in->end_pos) {
    return NULL;
  } else {
    if (in->buffer[in->end_posp-1] == '\r') {
      in->buffer[in->end_posp-1] = 0;
    } else {
      in->buffer[in->end_posp] = 0;
    }
    char *res = in->buffer+in->pos;
    in->pos = ++in->end_posp;
    return res;
  }
}
