// Larbin
// Sebastien Ailleret
// 03-02-00 -> 20-11-01

#include <iostream.h>
#include <string.h>
#include <unistd.h>

#include "options.h"

#include "types.h"
#include "global.h"
#include "fetch/file.h"
#include "utils/text.h"
#include "utils/debug.h"
#include "interf/output.h"

#ifdef SAVE

#ifndef MIRROR_SITES
static int nbdir = -1;
static int nbfile = filesPerDir;
static char buf[maxUrlSize+30];
#endif // MIRROR_SITES

static char *fileName;
static uint endFileName;
static void save (html *page);

#else // SAVE

#define save(x) ((void) 0)

#endif // SAVE

/** pageContent gives the content of the page
 * uses page->getPage usually
 * or read and delete temporary file if exists
 * the char * is always static
 *
 * deleteTemp is needed in endOfLoad for deleting temp file
 * if err != success */
#ifdef SPECIFICSAVE
static char *pageContent (html *page);
static void deleteTemp (html *page);
#else // SPECIFICSAVE not defined
#define pageContent(p) p->getPage()
#define deleteTemp(p) ((void) 0)
#endif // SPECIFICSAVE

/* This is the file you should change if you want to
 * use the data fetched by larbin.
 *
 * You should only change the 4 first functions :
 * loadedInteresting : an intersting  page has been loaded successfully
 *     this is only used with specificSearch
 * loaded : a page has been loaded successfully
 * fetchFailInteresting : a page has failed, but might be interesting
 *     this is only used with specificSearch
 * fetchFailStandard : fetch fail */


/** A page has been loaded successfully
 * @param page the page that has been fetched
 */
void loadedInteresting (html *page) {
  // Here should be the code for managing everything
  // page->getHeaders() gives a char* containing the http headers
  // pageContent(page) gives a char* containing the page itself
  // those char* are statically allocated, so you should copy
  // them if you want to keep them
#ifdef BIGSTATS
  cout << "fetched intersting : ";
  page->getUrl()->print();
#endif // BIGSTATS
}

/** A page has been loaded successfully
 * @param page the page that has been fetched
 */
void loaded (html *page) {
  // Here should be the code for managing everything
  // page->getHeaders() gives a char* containing the http headers
  // pageContent(page) gives a char* containing the page itself
  // those char* are statically allocated, so you should copy
  // them if you want to keep them
#ifdef BIGSTATS
  cout << "fetched : ";
  page->getUrl()->print();
  // cout << page->getHeaders() << "\n" << page->getPage() << "\n";
#endif // BIGSTATS
}

/** The fetch failed (but this page is interesting : only with specificSearch)
 * @param u the URL of the doc
 * @param reason reason of the fail
 */
void fetchFailInteresting (url *u, FetchError reason) {
  // Here should be the code for managing everything
#ifdef BIGSTATS
  cout << "fetched failed interesting (" << (int) reason << ") : ";
  u->print();
#endif // BIGSTATS
}

/** The fetch failed (but this page is interesting : only with specificSearch)
 * @param u the URL of the doc
 * @param reason reason of the fail
 */
void fetchFailStandard (url *u, FetchError reason) {
  // Here should be the code for managing everything
#ifdef BIGSTATS
  cout << "fetched failed (" << (int) reason << ") : ";
  u->print();
#endif // BIGSTATS
}


/**************************************************/
/* You should not change anything under this line */
/**************************************************/

/** The fetch failed
 * @param u the URL of the doc
 * @param reason reason of the fail
 */
void fetchFail (url *u, FetchError err, bool interesting=false) {
  // Here should be the code for managing everything
  if (global::isSpecific) {
    uint len = strlen(u->getFile());
    if (interesting
        || (global::privilegedExt != NULL
        && endWithIgnoreCase(global::privilegedExt, u->getFile(), len))) {
      fetchFailInteresting(u, err);
    } else {
      fetchFailStandard(u, err);
    }
  } else {
    fetchFailStandard(u, err);
  }
}

/** It's over with this file
 * report the situation ! (and make some stats)
 */
void endOfLoad (html *parser, FetchError err) {
  answers(err);
  switch (err) {
  case success:
	if (parser->isInteresting) {
	  interestingPage();
	  loadedInteresting(parser);
	} else {
      loaded(parser);
    }
    save(parser);
	break;
  default:
    fetchFail(parser->getUrl(), err, parser->isInteresting);
    // deleteTemp(parser); // uncomment for use of temp files
    break;
  }
}

#ifdef THREAD_OUTPUT
/** In this thread, end user manage the result of the crawl
 */
void *startOutput (void *none) {
  crash("Output on"); 
  for (;;) {
	Connexion *conn = global::userConns->get();
	endOfLoad((html *)conn->parser, conn->err);
	conn->recycle();
	global::freeConns->put(conn);
  }
  return NULL;
}
#endif

////////////////////////////////////////////////////////
/** this function gives the content of the page
 * uses page->getPage usually
 * or read and delete temporary file if exists
 * the char * is always static */
#ifdef SPECIFICSAVE
static char *pageContent (html *page) {
  static char content[maxSpecSize];
  if (page->mydir >= 0) {
    char *file = getSpecName(page->mydir, page->myfile);
    int fds = open (file, O_RDONLY);
    if (fds < 0) perror(file);
    int cont = 1;
    int pos = 0;
    while (cont) {
      cont = read(fds, content+pos, maxSpecSize-1-pos);
      pos += cont;
    }
    content[pos] = 0;
    close(fds);
    unlink(file);
    return content;
  } else {
    return page->getPage();
  }
}

static void deleteTemp (html *page) {
  if (page->mydir >= 0) {
    char *file = getSpecName(page->mydir, page->myfile);
    unlink(file);
  }
}
#endif // SPECIFICSAVE

////////////////////////////////////////////////////////
/** function for saving pages to files */
#ifdef SAVE

#ifdef MIRROR_SITES
#define endInitSave() \
  strcpy(fileName+endFileName, "d00000/"); \
  endFileName += 7 // indique le premier char a ecrire
#else
#define endInitSave() \
  strcpy(fileName+endFileName, "d00000/f00000"); \
  endFileName += 12 // indique le dernier char ecrit
#endif // MIRROR_SITES

void initSave () {
  mkdir(saveDir, S_IRWXU);
  endFileName = strlen(saveDir);
  fileName = new char[endFileName+maxUrlSize+50];
  strcpy(fileName, saveDir);
  if (fileName[endFileName-1] != '/') fileName[endFileName++] = '/';
  endInitSave();
}

#ifdef MIRROR_SITES

// save file
static void save (html *page) {
  // get file name and create needed directories
  url *u = page->getUrl();
  uint p = u->getPort();
  char *h = u->getHost();
  char *f = u->getFile();
  // update dir name
  uint d = u->hostHashCode() % nbDir;
  for (int i=2; i<7; i++) {
    fileName[endFileName-i] = d % 10 + '0';
    d /= 10;
  }
  // set file name
  uint len = endFileName;
  if (p == 80)
    len += sprintf(fileName+endFileName, "%s%s", h, f);
  else
    len += sprintf(fileName+endFileName, "%s:%u%s", h, p, f);
  // make sure the path of the file exists
  bool cont = true;
  struct stat st;
  while (cont) {
    len--;
    while (fileName[len] != '/') len--;
    fileName[len] = 0;
    cont = stat(fileName, &st); // this becomes true at least for saveDir
    fileName[len] = '/';
  }
  cont = true;
  while (cont) {
    len++;
    while (fileName[len] != '/' && fileName[len] != 0) len++;
    if (fileName[len] == '/') {
      fileName[len] = 0;
      if (mkdir(fileName, S_IRWXU) != 0) perror("trouble while creating dir");
      fileName[len] = '/';
    } else { // fileName[len] == 0
      cont = false;
    }
  }
  if (fileName[len-1] == '/') {
    strcpy(fileName+len, indexFile);
  }
  // open fds and write file
  int fd = creat(fileName, S_IRWXU);
  if (fd >= 0) {
    // some url mysteries might prevent this from being possible
    // ex: if http://bbs.computoredge.com/ceo
    // and http://bbs.computoredge.com/ceo/ both exist
    ecrire(fd, page->getPage());
    close(fd);
  }
}  

#else // MIRROR_SITES not defined

// save file
static void save (html *page) {
  nbfile++;
  if (nbfile < filesPerDir) {
    int tmp = nbfile;
    int pos = endFileName;
    while (tmp != 0) {
      fileName[pos--] = '0' + (tmp % 10);
      tmp /= 10;
    }
  } else { // new dir
    nbdir++; nbfile = 0;
    for (uint i=endFileName; i>endFileName-5; i--) fileName[i]='0';
    int pos = endFileName-7;
    int tmp = nbdir;
    while (tmp != 0) {
      fileName[pos--] = '0' + (tmp % 10);
      tmp /= 10;
    }
    // create the directory
    fileName[endFileName-6] = 0;
    if (mkdir(fileName, S_IRWXU) != 0) perror("trouble while creating dir");
    fileName[endFileName-6] = '/';
  }
  int fd = creat(fileName, S_IRWXU);
  if (fd < 0) {
    cerr << "cannot open file " << fileName << "\n";
    exit(1);
  }
  int s=0;
#ifdef URL_TAGS
  s = sprintf(buf, "%u ", page->getUrl()->tag);
#endif // URL_TAGS
  s += page->getUrl()->writeUrl(buf+s);
  buf[s++] = '\n';
  buf[s++] = '\n';
  int len = strlen(page->getHeaders());
  if (len+1 >= s) {
    char *tmp = page->getPage() - s;
    memmove(tmp, buf, s);
    ecrire(fd, tmp);
  } else {
    ecrireBuff(fd, buf, s);
    ecrire(fd, page->getPage());
  }
  close(fd);
}
#endif // MIRROR_SITES

#endif // SAVE
