// Larbin
// Sebastien Ailleret
// 09-11-99 -> 20-11-01

#include <unistd.h>
#include <iostream.h>
#include <signal.h>
#include <time.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>

#include "options.h"

#include "global.h"
#include "utils/url.h"
#include "utils/text.h"
#include "fetch/checker.h"
#include "fetch/sequencer.h"
#include "fetch/fetchOpen.h"
#include "fetch/fetchPipe.h"
#include "interf/input.h"
#include "interf/output.h"

#include "utils/debug.h"
#include "utils/webserver.h"
#include "utils/histogram.h"

typedef void* (*StartFun) (void *);

static void cron ();
static void startThread (StartFun run, void *arg);

///////////////////////////////////////////////////////////
#ifdef PROF
static bool stop = false;
static void handler (int i) {
  stop = true;
}
#endif // PROF

// wait to limit bandwidth usage
#ifdef MAXBANDWIDTH
static void waitBandwidth (time_t *old) {
  while (global::remainBand < 0) {
    poll(NULL, 0, 10);
    global::now = time(NULL);
    if (*old != global::now) {
      *old = global::now;
      cron ();
    }
  }
}
#else
#define waitBandwidth(x) ((void) 0)
#endif // MAXBANDWIDTH

#ifndef NDEBUG
static uint count = 0;
#endif // NDEBUG

///////////////////////////////////////////////////////////
// If this thread terminates, the whole program exits
int main (int argc, char *argv[]) {
  // create all the structures
  global glob(argc, argv);
#ifdef PROF
  signal (2, handler);
#endif // PROF

  // Start all threads (interaction with the user)
  crash("Launch all threads");
#ifdef THREAD_OUTPUT
  startThread(startOutput, NULL);
#endif
  startThread(startWebserver, NULL);

  // Start the search
  cerr << global::userAgent;
  cerr << " is starting its search\n";
  if (global::firstUrl != NULL) {
#ifdef URL_TAGS
    check(new url(global::firstUrl, global::depthInSite, 0));
#else
    check(new url(global::firstUrl, global::depthInSite));
#endif // URL_TAGS
  }

  time_t old = global::now;
  for (;;) {
    // update time
    global::now = time(NULL);
    if (old != global::now) {
      // this block is called every second
      old = global::now;
      cron();
    }
    stateMain(-count);
    waitBandwidth(&old);
    stateMain(1);
    for (int i=0; i<global::maxFds; i++)
      global::ansPoll[i] = 0;
    for (uint i=0; i<global::posPoll; i++)
      global::ansPoll[global::pollfds[i].fd] = global::pollfds[i].revents;
    global::posPoll = 0;
    stateMain(2);
    input();
    stateMain(3);
    sequencer();
    stateMain(4);
    fetchDns();
    stateMain(5);
    fetchOpen();
    stateMain(6);
    checkAll();
    // select
    stateMain(count++);
    poll(global::pollfds, global::posPoll, 10);
    stateMain(7);
  }
}

// a lot of stats and profiling things
static void cron () {
  // shall we stop
#ifdef PROF
  if (stop) exit(2);
#endif // PROF
#ifdef EXIT_AT_END
  if (global::URLsInternal->getLength() == 0
      && debUrl == 0)
    exit(0);
#endif // EXIT_AT_END

  // look for timeouts
  checkTimeout();

#ifdef MAXBANDWIDTH
  // give bandwidth
  if (global::remainBand > 0) {
    global::remainBand = MAXBANDWIDTH;
  } else {
    global::remainBand = global::remainBand + MAXBANDWIDTH;
  }
#endif // MAXBANDWIDTH

#ifndef NOSTATS
#ifdef GRAPH
  histoHit(pages, answers[success]);
#endif // GRAPH

  if ((global::now & 7) == 0) {
    urlsRate = (urls - urlsPrev) >> 3; urlsPrev = urls;
    pagesRate = (pages - pagesPrev) >> 3; pagesPrev = pages;
    successRate = (answers[success] - successPrev) >> 3;
    successPrev = answers[success];
    siteSeenRate = (siteSeen - siteSeenPrev) >> 3; siteSeenPrev = siteSeen;
    siteDNSRate = (siteDNS - siteDNSPrev) >> 3; siteDNSPrev = siteDNS;
#ifndef NDEBUG
    readRate = (byte_read - readPrev) >> 3; readPrev = byte_read;
    writeRate = (byte_write - writePrev) >> 3; writePrev = byte_write;
#endif // NDEBUG

#ifdef STATS
    cout << "\n" << ctime(&global::now)
         << "urls : " << urls << "  (rate : " << urlsRate
         << ")\npages : " << pages << "  (rate : " << pagesRate
         << ")\nsuccess : " << answers[success]
         << "  (rate : " << successRate << ")\n";
#endif // STATS
  }
#endif // NOSTATS
}

/* Launch a new thread
 * return 0 in case of success
 */
static void startThread (StartFun run, void *arg) {
  pthread_t t;
  pthread_attr_t attr;
  if (pthread_attr_init(&attr) != 0
      || pthread_create(&t, &attr, run, arg) != 0
      || pthread_attr_destroy(&attr) != 0
      || pthread_detach(t) != 0) {
    cerr << "Unable to launch a thread\n";
    exit(1);
  }
}
