// Larbin
// Sebastien Ailleret
// 15-11-99 -> 22-07-01

/* This class describes an URL */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <iostream.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "options.h"

#include "types.h"
#include "global.h"
#include "utils/url.h"
#include "utils/text.h"
#include "utils/connexion.h"
#include "utils/debug.h"

uint siteHashCode (char *host, int port) {
  uint h=port;
  uint i;
  for (i=0; i < strlen(host); i++) {
	h = 37*h + host[i];
  }
  return h % namedSiteListSize;
}

uint int_of_hexa (char c) {
  if (c >= '0' && c <= '9')
    return (c - '0');
  else if (c >= 'a' && c <= 'f')
    return (c - 'a' + 10);
  else if (c >= 'A' && c <= 'F')
    return (c - 'A' + 10);
  else assert(false);
  // never reached
  return 0;
}

/* Constructor : Parses an url */
#ifdef URL_TAGS
url::url (char *u, uint depth, uint tag, url *base) {
  this->tag = tag;
#else
url::url (char *u, uint depth, url *base) {
#endif // URL_TAGS
  newUrl();
  this->depth = depth;
  host = NULL;
  port = 80;
  file = NULL;
  if (startWithIgnoreCase("http://", u)) {
	// absolute url
	parse (u + 7);
    // normalize file name
    if (file != NULL && !normalize(file)) {
      delete [] file;
      file = NULL;
      delete [] host;
      host = NULL;
    }
  } else if (base != NULL) {
	if (startWithIgnoreCase("http:", u)) {
	  parseWithBase(u+5, base);
	} else if (isProtocol(u)) {
	  // Unknown protocol (mailto, ftp, news, file, gopher...)
	} else {
	  parseWithBase(u, base);
	}
  }
}

/* Constructor : read the url from a file (cf serialize)
 */
url::url (char *line) {
  newUrl();
  int i=0;
  // Read depth
  depth = 0;
  while (line[i] >= '0' && line[i] <= '9') {
	depth = 10*depth + line[i] - '0';
	i++;
  }
#ifdef URL_TAGS
  // read tag
  tag = 0; i++;
  while (line[i] >= '0' && line[i] <= '9') {
	tag = 10*tag + line[i] - '0';
	i++;
  }
#endif // URL_TAGS
  int deb = ++i;
  // Read host
  while (line[i] != ':') {
	i++;
  }
  line[i] = 0;
  host = newString(line+deb);
  i++;
  // Read port
  port = 0;
  while (line[i] >= '0' && line[i] <= '9') {
	port = 10*port + line[i] - '0';
	i++;
  }
  // Read file name
  file = newString(line+i);
}
 
url::url (char *host, uint port, char *file) {
  newUrl();
  this->host = host;
  this->port = port;
  this->file = file;
}
   
/* Destructor */
url::~url () {
  delUrl();
  delete [] host;
  delete [] file;
}

/* Is it a valid url ? */
bool url::isValid () {
  return host!=NULL && file!=NULL
    && strlen(host) + strlen(file) + 18 < maxUrlSize;
}

/* print an URL */
void url::print () {
  cout << "http://" << host << ":" << port << file << "\n";
}

/** Set depth to one if we are at an entry point in the site */
void url::setDepth (char *fromHost) {
  if (strcmp(fromHost, host)) {
	depth = global::depthInSite;
  }
}

/* return the base of the url */
url *url::giveBase () {
  int i = strlen(file);
  assert (file[0] == '/');
  while (file[i] != '/') {
	i--;
  }
  char *newFile = new char[i+2];
  memcpy(newFile, file, i+1);
  newFile[i+1] = 0;
  return new url(newString(host), port, newFile);
}

/** return a char * representation of the url
 * give means that you have to delete the string yourself
 */
char *url::giveUrl () {
  char *tmp;
  int i = strlen(file);
  int j = strlen(host);

  tmp = new char[18+i+j];  // 7 + j + 1 + 9 + i + 1
                           // http://(host):(port)(file)\0
  strcpy(tmp, "http://");
  strcpy (tmp+7, host);
  j += 7;
  if (port != 80) {
    j += sprintf(tmp + j, ":%u", port);
  }
  // Copy file name
  while (i >= 0) {
	tmp [j+i] = file[i];
	i--;
  }
  return tmp;
}

/** write the url in a buffer
 *  buf must be at least of size maxUrlSize
 * returns the size of what has been written (not including '\0')
 */
int url::writeUrl (char *buf) {
  if (port == 80)
    return sprintf(buf, "http://%s%s", host, file);
  else
    return sprintf(buf, "http://%s:%u%s", host, port, file);
}

/* return a hashcode for the host of this url */
uint url::hostHashCode () {
  return siteHashCode (host, port);
}

/* return a hashcode for this url */
uint url::hashCode () {
  unsigned int h=port;
  unsigned int i;
  for (i=0; i < strlen(host); i++) {
	h = 31*h + host[i];
  }
  for (i=0; i < strlen(file); i++) {
	h = 31*h + file[i];
  }
  return h % hashSize;
}

/* parses a url : 
 * at the end, arg must have its initial state, 
 * http:// has allready been suppressed
 */
void url::parse (char *arg) {
  int deb = 0, fin = deb;
  // Find the end of host name (put it into lowerCase)
  while (arg[fin] != '/' && arg[fin] != ':' && arg[fin] != 0) {
	fin++;
  }
  if (fin == 0) return;

  // get host name
  host = new char[fin+1];
  for (int  i=0; i<fin; i++) {
    host[i] = tolower(arg[i]);
  }
  host[fin] = 0;

  // get port number
  if (arg[fin] == ':') {
	port = 0;
    fin++;
	while (arg[fin] >= '0' && arg[fin] <= '9') {
	  port = port*10 + arg[fin]-'0';
	  fin++;
	}
  }

  // get file name
  if (arg[fin] != '/') {
	// www.inria.fr => add the final /
    file = newString("/");
  } else {
    file = newString(arg + fin);
  }
}

/** parse a file with base
 */
void url::parseWithBase (char *u, url *base) {
  // cat filebase and file
  if (u[0] == '/') {
    file = newString(u);
  } else {
    uint lenb = strlen(base->file);
    char *tmp = new char[lenb + strlen(u) + 1];
    memcpy(tmp, base->file, lenb);
    strcpy(tmp + lenb, u);
    file = tmp;
  }
  if (!normalize(file)) {
    delete [] file;
    file = NULL;
    return;
  }
  host = newString(base->host);
  port = base->port;
}

/** normalize file name
 * @return true if it is ok, false otherwise (cgi-bin)
 */
bool url::normalize (char *file) {
  int i=0;
  while (file[i] != 0 && file[i] != '#') {
	if (file[i] == '/') {
	  if (file[i+1] == '.' && file[i+2] == '/') {
		// suppress /./
		int j=i+3;
		while (file[j] != 0) {
		  file[j-2] = file[j];
		  j++;
		}
		file[j-2] = 0;
      } else if (file[i+1] == '/') {
        // replace // by /
        int j=i+2;
        while (file[j] != 0) {
          file[j-1] = file[j];
		  j++;
        }
        file[j-1] = 0;
      } else if (file[i+1] == '.' && file[i+2] == '.' && file[i+3] == '/') {
		// suppress /../
		if (i == 0) {
		  // the file name starts with /../ : error
		  return false;
		} else {
		  int j = i+4, dec;
		  i--;
		  while (file[i] != '/') { i--; }
		  dec = i+1-j; // dec < 0
		  while (file[j] != 0) {
			file[j+dec] = file[j];
			j++;
		  }
		  file[j+dec] = 0;
		}
	  } else if (file[i+1] == '.' && file[i+2] == 0) {
		// suppress /.
        file[i+1] = 0;
        return true;
	  } else if (file[i+1] == '.' && file[i+2] == '.' && file[i+3] == 0) {
		// suppress /..
		if (i == 0) {
          // the file name starts with /.. : error
		  return false;
		} else {
		  i--;
		  while (file[i] != '/') {
			i--;
		  }
          file[i+1] = 0;
          return true;
		}
	  } else { // nothing special, go forward
		i++;
	  }
	} else if (file[i] == '%') {
      if (!isxdigit(file[i+1]) || !isxdigit(file[i+2])) return false;
      char c = 16 * int_of_hexa(file[i+1]) + int_of_hexa(file[i+2]);
      if (isgraph(c)) {
        file[i] = c;
        int j = i+3;
        while (file[j] != 0) {
          file[j-2] = file[j];
          j++;
        }
        file[j-2] = 0;
        i++;
      } else if (c == ' ') { // keep it with the % notation
        i += 3;
      } else { // bad url
        return false;
      }
    } else if (file[i] == 'c' && file[i+1] == 'g' && file[i+2] == 'i') {
      // avoid many realmedia redirection pages :
      // http://..../Realmedia/bla.cgi/...
      return false;
    } else { // nothing special, go forward
	  i++;
	}
  }
  file[i] = 0;
  return true;
}

/* Does this url starts with a protocol name */
bool url::isProtocol (char *s) {
  uint i = 0;
  while (isalnum(s[i])) {
	i++;
  }
  return s[i] == ':';
}

/* serialize the url */
char *url::serialize () {
  // this buffer is protected by the lock of PersFifo
  static char s[maxUrlSize+40];
#ifdef URL_TAGS
  sprintf(s, "%u %u %s:%u%s\n", depth, tag, host, port, file);
#else
  sprintf(s, "%u %s:%u%s\n", depth, host, port, file);
#endif // URL_TAGS
  return s;
}
