// Larbin
// Sebastien Ailleret
// 15-11-99 -> 30-05-01

/* This class describes an URL */

#ifndef URL_H
#define URL_H

#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>

#include "types.h"
#include "utils/string.h"

uint siteHashCode (char *host, int port);

struct global;

class url {
 private:
  char *host;
  uint port;
  char *file;
  int depth;
  /* parse the url */
  void parse (char *s);
  /** parse a file with base */
  void parseWithBase (char *u, url *base);
  /* normalize file name */
  bool normalize (char *file);
  /* Does this url starts with a protocol name */
  bool isProtocol (char *s);

 public:
  /* Constructor : Parses an url (u is deleted) */
#ifdef URL_TAGS
  url (char *u, uint depth, uint tag, url *base = NULL);
#else
  url (char *u, uint depth, url *base = NULL);
#endif

  /* Constructor : read the url from a file (cf serialize) */
  url (char *line);

  /* constructor used by giveBase */
  url (char *host, uint port, char *file);

  /* Destructor */
  ~url ();

  /* inet addr (once calculated) */
  sockaddr_in addr;

  /* Is it a valid url ? */
  bool isValid ();

  /* print an URL */
  void print ();

  /* return the host */
  inline char *getHost () { return host; }

  /* return the port */
  inline uint getPort () { return port; }

  /* return the file */
  inline char *getFile () { return file; }

  /** Depth in the Site */
  inline int getDepth () { return depth; }

  /** Set depth to one if we are at an entry point in the site */
  void setDepth (char *fromHost);

  /** return the base of the url
   * give means that you have to delete the string yourself
   */
  url *giveBase ();

  /** return a char * representation of the url
   * give means that you have to delete the string yourself
   */
  char *giveUrl ();

  /** write the url in a buffer
   *  buf must be at least of size maxUrlSize
   * returns the size of what has been written (not including '\0')
   */
  int writeUrl (char *buf);

  /* return a hashcode for the host of this url */
  uint hostHashCode ();

  /* return a hashcode for this url */
  uint hashCode ();

  /* serialize the url */
  char *serialize ();

#ifdef URL_TAGS
  /* tag associated to this url */
  uint tag;
#endif // URL_TAGS
};

#endif // URL_H
